import ctypes as _c
from . import _read_string
from . import _lls
from . import _ls_check_errors
from . import LSError
from . import LSErrorCode
from . import LSExpression
from . import _encode_string
from . import _is_string
import numbers
import sys
from . import LocalSolver
from . import _encode_string_no_null
from . import _decode_subset_bytes
from . import _ls_interrupt
from . import _ls_set_pending_error
from . import _ls_define_signal_handler
from . import _encode_strings_no_null

########


__all__ = [
    "LSPModeler", "LSPModule", "LSPValue", "LSPFunction", "LSPMap", "LSPReferenceScope"
]


########


class _lspdata_val(_c.Union):
    _fields_ = [('ref', _c.c_void_p),
                ('intValue', _c.c_longlong),
                ('dblValue', _c.c_double)]


class _lspdata(_c.Structure):
    _anonymous_ = ('val',)
    _fields_ = [('val', _lspdata_val),
                ('type', _c.c_size_t)]

class _lsp_ref_guard(object):
    __slots__ = "_ref",

    def __init__(self, ref):
        self._ref = ref

    def __enter__(self):
        _lsp_inc_ref(self._ref)

    def __exit__(self, _1, _2, _3):
        _lsp_dec_ref(self._ref)

_lsp_create_modeler = _lls.lsp_create_modeler
_lsp_create_modeler.argtypes = []
_lsp_create_modeler.restype = _c.c_void_p
_lsp_create_modeler.errcheck = _ls_check_errors

_lsp_delete_modeler = _lls.lsp_delete_modeler
_lsp_delete_modeler.argtypes = [_c.c_void_p]
_lsp_delete_modeler.restype = None
_lsp_delete_modeler.errcheck = _ls_check_errors

_lsp_stream_writer_type = _c.CFUNCTYPE(None, _c.c_void_p, _c.c_char_p, _c.c_int, _c.c_void_p)
_lsp_stream_flusher_type = _c.CFUNCTYPE(None, _c.c_void_p, _c.c_void_p)
_lsp_set_std_stream = _lls.lsp_set_std_stream
_lsp_set_std_stream.argtypes = [_c.c_void_p, _c.c_int, _lsp_stream_writer_type, _lsp_stream_flusher_type, _c.c_void_p]
_lsp_set_std_stream.restype = None
_lsp_set_std_stream.errcheck = _ls_check_errors


_lsp_create_module = _lls.lsp_create_module
_lsp_create_module.argtypes = [_c.c_void_p, _c.c_char_p]
_lsp_create_module.restype = _lspdata
_lsp_create_module.errcheck = _ls_check_errors

_lsp_load_module = _lls.lsp_load_module
_lsp_load_module.argtypes = [_c.c_void_p, _c.c_char_p]
_lsp_load_module.restype = _lspdata
_lsp_load_module.errcheck = _ls_check_errors

_lsp_module_get = _lls.lsp_module_get
_lsp_module_get.argtypes = [_c.c_void_p, _c.c_char_p]
_lsp_module_get.restype = _lspdata
_lsp_module_get.errcheck = _ls_check_errors

_lsp_module_set = _lls.lsp_module_set
_lsp_module_set.argtypes = [_c.c_void_p, _c.c_char_p, _lspdata]
_lsp_module_set.restype = None
_lsp_module_set.errcheck = _ls_check_errors

_lsp_module_run = _lls.lsp_module_run
_lsp_module_run.argtypes = [_c.c_void_p, _c.c_void_p, _c.c_char_p, _c.c_int]
_lsp_module_run.restype = None
_lsp_module_run.errcheck = _ls_check_errors

_lsp_module_run_main = _lls.lsp_module_run_main
_lsp_module_run_main.argtypes = [_c.c_void_p, _c.c_char_p, _c.c_int]
_lsp_module_run_main.restype = None
_lsp_module_run_main.errcheck = _ls_check_errors


_lsp_create_map = _lls.lsp_create_map
_lsp_create_map.argtypes = [_c.c_void_p]
_lsp_create_map.restype = _lspdata
_lsp_create_map.errcheck = _ls_check_errors

_lsp_map_is_defined = _lls.lsp_map_is_defined
_lsp_map_is_defined.argtypes = [_c.c_void_p, _lspdata]
_lsp_map_is_defined.restype = _c.c_bool
_lsp_map_is_defined.errcheck = _ls_check_errors

_lsp_map_get = _lls.lsp_map_get
_lsp_map_get.argtypes = [_c.c_void_p, _lspdata]
_lsp_map_get.restype = _lspdata
_lsp_map_get.errcheck = _ls_check_errors

_lsp_map_set = _lls.lsp_map_set
_lsp_map_set.argtypes = [_c.c_void_p, _lspdata, _lspdata]
_lsp_map_set.restype = None
_lsp_map_set.errcheck = _ls_check_errors

_lsp_map_add = _lls.lsp_map_add
_lsp_map_add.argtypes = [_c.c_void_p, _lspdata]
_lsp_map_add.restype = None
_lsp_map_add.errcheck = _ls_check_errors

_lsp_map_clear = _lls.lsp_map_clear
_lsp_map_clear.argtypes = [_c.c_void_p]
_lsp_map_clear.restype = None
_lsp_map_clear.errcheck = _ls_check_errors

_lsp_map_count = _lls.lsp_map_count
_lsp_map_count.argtypes = [_c.c_void_p]
_lsp_map_count.restype = _c.c_longlong
_lsp_map_count.errcheck = _ls_check_errors


_lsp_create_expr = _lls.lsp_create_expr
_lsp_create_expr.argtypes = [_c.c_void_p, _c.c_int]
_lsp_create_expr.restype = _lspdata
_lsp_create_expr.errcheck = _ls_check_errors

_lsp_expr_index = _lls.lsp_expr_index
_lsp_expr_index.argtypes = [_c.c_void_p]
_lsp_expr_index.restype = _c.c_int
_lsp_expr_index.errcheck = _ls_check_errors


_lsp_create_string = _lls.lsp_create_string
_lsp_create_string.argtypes = [_c.c_void_p, _c.c_char_p]
_lsp_create_string.restype = _lspdata
_lsp_create_string.errcheck = _ls_check_errors

_lsp_create_string_2 = _lls.lsp_create_string
_lsp_create_string_2.argtypes = [_c.c_void_p, _c.c_char_p, _c.c_int]
_lsp_create_string_2.restype = _lspdata
_lsp_create_string_2.errcheck = _ls_check_errors

_lsp_string = _lls.lsp_string
_lsp_string.argtypes = [_c.c_void_p, _c.c_char_p, _c.c_int]
_lsp_string.restype = _c.c_int
_lsp_string.errcheck = _ls_check_errors

_lsp_function_type_2 = _c.CFUNCTYPE(
    None,
    _c.c_void_p,
    _c.POINTER(_lspdata),
    _c.c_int,
    _c.POINTER(_lspdata),
    _c.c_void_p
)

_lsp_create_function_2 = _lls.lsp_create_function_2
_lsp_create_function_2.argtypes = [_c.c_void_p, _c.c_char_p, _lsp_function_type_2, _c.c_void_p]
_lsp_create_function_2.restype = _lspdata
_lsp_create_function_2.errcheck = _ls_check_errors

_lsp_function_call = _lls.lsp_function_call
_lsp_function_call.argtypes = [_c.c_void_p, _c.POINTER(_lspdata), _c.c_int]
_lsp_function_call.restype = _lspdata
_lsp_function_call.errcheck = _ls_check_errors

_lsp_function_name = _lls.lsp_function_name
_lsp_function_name.argtypes = [_c.c_void_p, _c.c_char_p, _c.c_int]
_lsp_function_name.restype = _c.c_int
_lsp_function_name.errcheck = _ls_check_errors


_lsp_create_iterator = _lls.lsp_create_iterator
_lsp_create_iterator.argtypes = [_c.c_void_p]
_lsp_create_iterator.restype = _lspdata
_lsp_create_iterator.errcheck = _ls_check_errors

_lsp_iterator_next = _lls.lsp_iterator_next
_lsp_iterator_next.argtypes = [_c.c_void_p, _c.POINTER(_lspdata), _c.POINTER(_lspdata)]
_lsp_iterator_next.restype = _c.c_bool
_lsp_iterator_next.errcheck = _ls_check_errors


_lsp_inc_ref = _lls.lsp_inc_ref
_lsp_inc_ref.argtypes = [_c.c_void_p]
_lsp_inc_ref.restype = None
_lsp_inc_ref.errcheck = _ls_check_errors

_lsp_dec_ref = _lls.lsp_dec_ref
_lsp_dec_ref.argtypes = [_c.c_void_p]
_lsp_dec_ref.restype = None
_lsp_dec_ref.errcheck = _ls_check_errors

_lsp_type = _lls.lsp_type
_lsp_type.argtypes = [_c.c_size_t]
_lsp_type.restype = _c.c_int
_lsp_type.errcheck = _ls_check_errors


_lsp_check_type = _lls.lsp_check_type
_lsp_check_type.argtypes = [_c.c_size_t, _c.c_int]
_lsp_check_type.restype = _c.c_bool
_lsp_check_type.errcheck = _ls_check_errors


_lsp_solver = _lls.lsp_solver
_lsp_solver.argtypes = [_c.c_void_p]
_lsp_solver.restype = _c.c_void_p
_lsp_solver.errcheck = _ls_check_errors

_lsp_create_solver = _lls.lsp_create_solver
_lsp_create_solver.argtypes = [_c.c_void_p];
_lsp_create_solver.restype = _lspdata
_lsp_create_solver.errcheck = _ls_check_errors

_lsp_solver_handle = _lls.lsp_solver_handle
_lsp_solver_handle.argtypes = [_c.c_void_p];
_lsp_solver_handle.restype = _lspdata
_lsp_solver_handle.errcheck = _ls_check_errors

_lsp_solver_reset = _lls.lsp_solver_reset
_lsp_solver_reset.argtypes = [_c.c_void_p];
_lsp_solver_reset.restype = None
_lsp_solver_reset.errcheck = _ls_check_errors

_lsp_create_expr = _lls.lsp_create_expr
_lsp_create_expr.argtypes = [_c.c_void_p, _c.c_int];
_lsp_create_expr.restype = _lspdata
_lsp_create_expr.errcheck = _ls_check_errors

_lsp_expr_handle = _lls.lsp_expr_handle
_lsp_expr_handle.argtypes = [_c.c_void_p];
_lsp_expr_handle.restype = _lspdata
_lsp_expr_handle.errcheck = _ls_check_errors

_lsp_expr_index = _lls.lsp_expr_index
_lsp_expr_index.argtypes = [_c.c_void_p];
_lsp_expr_index.restype = _c.c_int
_lsp_expr_index.errcheck = _ls_check_errors

_lsp_handle_solver = _lls.lsp_handle_solver
_lsp_handle_solver.argtypes = [_c.c_void_p];
_lsp_handle_solver.restype = _c.c_void_p
_lsp_handle_solver.errcheck = _ls_check_errors



########



class LSPValue(object):
    __slots__ = "_modeler", "_data", "_ref"

    def __init__(self, modeler, data):
        self._modeler = modeler
        self._data = data
        self._ref = data.val.ref
        _lsp_inc_ref(self._ref)
        modeler._reference_scopes[-1]._references.append(self)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        _check_modeler(self._modeler)
        self.__del__()

    def __del__(self):
        if self._ref != 0 and self._modeler._modeler_ptr is not None:
            _lsp_dec_ref(self._ref)
            self._data = 0
            self._ref = 0

    def __eq__(self, other):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        return (isinstance(other, LSPValue)
                and self._ref == other._ref)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        return self._ref

########


def _create_lsp_data(modeler, value):
    if value is None:
        return _lspdata()
    elif isinstance(value, bool):
        result = _lspdata()
        result.val.intValue = value
        result.type = 5
        return result
    elif isinstance(value, int):
        result = _lspdata()
        result.val.intValue = value
        result.type = 1
        return result
    elif isinstance(value, float):
        result = _lspdata()
        result.val.dblValue = value
        result.type = 2
        return result
    elif isinstance(value, LSExpression):
        _check_attached_solver(modeler, value._solver)
        return _lsp_create_expr(value._solver._handle_modeler_ptr, value.index)
    elif _is_string(value):
        raw_str = _encode_string(value)
        return _lsp_create_string_2(modeler._modeler_ptr, raw_str, len(raw_str))
    elif (isinstance(value, LSPValue)):
        return value._data
    elif isinstance(value, numbers.Integral):
        result = _lspdata()
        result.val.intValue = int(value)
        result.type = 1
        return result
    elif isinstance(value, numbers.Real):
        result = _lspdata()
        result.val.dblValue = float(value)
        result.type = 2
        return result
    else:
        raise TypeError("Cannot convert value of type '{}' to a valid modeler type."
                        .format(type(value)))

def _extract_lsp_data(modeler, data):
    if data.type == 0:
        return None
    elif data.type == 1:
        return data.val.intValue
    elif data.type == 2:
        return data.val.dblValue
    elif data.type == 5:
        if data.val.intValue == 1: return True
        else: return False

    value_type = _lsp_type(data.type)
    if value_type == 3:
        return _read_string(lambda buf, x: _lsp_string(data.ref, buf, x))
    elif value_type == 4:
        handle = _lsp_expr_handle(data.ref)
        expr_id = _lsp_expr_index(data.ref)
        ls = _retrieve_localsolver_from_handle(modeler, handle.ref)
        return LSExpression(ls, expr_id)
    elif value_type == 5:
        return LSPFunction(modeler, data)
    elif value_type == 6:
        return LSPMap(modeler, data)
    elif value_type == 7:
        return LSPModule(modeler, data)
    else:
        return LSPValue(modeler, data)


def _check_modeler(modeler):
    if modeler._modeler_ptr is None:
        raise LSError(LSErrorCode.API, "Cannot perform the asked operation on a deleted environment.",
                "lsputils.py", "_check_modeler", -1)

def _check_attached_solver(modeler, ls):
    if ls._handle_modeler_ptr is None or ls not in modeler._attached_solvers:
        raise LSError(LSErrorCode.API,
                "The solver is not associated to a modeler. Create your solver with the method LSPModeler.create_solver() instead",
                "lsputils.py", "_check_attached_solver", -1)

def _check_ref(reference):
    if reference == 0:
        raise LSError(LSErrorCode.API, "Cannot call this method on a disposed object.",
                "lsputils.py", "_check_ref", -1)


def _retrieve_localsolver_from_handle(modeler, handle):
    for ls in modeler._attached_solvers:
        if ls._handle_modeler_ptr == handle:
            return ls
    
    raise LSError(LSErrorCode.API,
            "Unable to retrieve LSExpressions based on a LocalSolver model that was not created in the Python API.",
            "lsputils.py", "_retrieve_localsolver_from_handle", -1)


########




class LSPFunction(LSPValue):
    __slots__ = "_args_array",

    def __init__(self, modeler, data):
        super(LSPFunction, self).__init__(modeler, data)
        self._args_array = None

    def get_name(self):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        return _read_string(lambda buf, x: _lsp_function_name(self._ref, buf, x))

    def __call__(self, *args):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        if self._args_array is None or len(self._args_array) < len(args) + 1:
            self._args_array = (_lspdata * (len(args) + 1))()

        for i, arg in enumerate(args, start=1):
            self._args_array[i] = _create_lsp_data(self._modeler, arg)

        result = _lsp_function_call(self._ref, self._args_array, len(args) + 1)
        return _extract_lsp_data(self._modeler, result)

    name = property(get_name)
########


class LSPMap(LSPValue):
    __slots__ = ()

    def __init__(self, modeler, data):
        super(LSPMap, self).__init__(modeler, data)

    def __len__(self):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        return _lsp_map_count(self._ref)

    def __getitem__(self, key):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_key = _create_lsp_data(self._modeler, key)
        lsp_val = _lsp_map_get(self._ref, lsp_key)
        return _extract_lsp_data(self._modeler, lsp_val)

    def __setitem__(self, key, value):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_key = _create_lsp_data(self._modeler, key)
        lsp_val = _create_lsp_data(self._modeler, value)
        _lsp_map_set(self._ref, lsp_key, lsp_val)

    def __delitem__(self, key):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_key = _create_lsp_data(self._modeler, key)
        _lsp_map_set(self._ref, lsp_key, _lspdata())

    def __contains__(self, key):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_key = _create_lsp_data(self._modeler, key)
        return _lsp_map_is_defined(self._ref, lsp_key)

    def __iter__(self):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        iterator_ref = _lsp_create_iterator(self._ref).val.ref
        with _lsp_ref_guard(iterator_ref):
            key_data = _lspdata()
            value_data = _lspdata()
            while _lsp_iterator_next(iterator_ref, _c.byref(key_data), _c.byref(value_data)):
                yield _extract_lsp_data(self._modeler, key_data), _extract_lsp_data(self._modeler, value_data)
                _check_modeler(self._modeler)
                _check_ref(self._ref)
                key_data = _lspdata()
                value_data = _lspdata()

    def clear(self):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        return _lsp_map_clear(self._ref)

    def append(self, value):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_val = _create_lsp_data(self._modeler, value)
        _lsp_map_add(self._ref, lsp_val)

    def get(self, key, default = None):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_key = _create_lsp_data(self._modeler, key)
        lsp_val = _lsp_map_get(self._ref, lsp_key)
        if lsp_val.type == 0: return default
        return _extract_lsp_data(self._modeler, lsp_val)




########


class LSPReferenceScope(object):
    __slots__ = "_modeler", "_references"

    def __init__(self, modeler):
        self._modeler = modeler
        self._references = []
        modeler._reference_scopes.append(self)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        _check_modeler(self._modeler)
        self.__del__()

    def __del__(self):
        if self._modeler._modeler_ptr is None:
            return

        try:
            self._modeler._reference_scopes.remove(self)
        except ValueError:
            pass

        for ref in self._references:
            ref.__exit__()
        self._references = []



########


class LSPModule(LSPValue):
    __slots__ = ()

    def __init__(self, modeler, data):
        super(LSPModule, self).__init__(modeler, data)

    def parse_arguments(self, *args):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        for arg in args:
            _lsp_module_parse_argument(self._ref, _encode_string_no_null(arg))

    def run(self, solver, *args):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        _check_attached_solver(self._modeler, solver)
        # Required since Jupyter overwrites the signal handler previously set without chaining it.
        _ls_define_signal_handler("SIGINT")
        str_args = _encode_strings_no_null(args)
        _lsp_module_run(self._ref, solver._solver_modeler_ptr, str_args, len(str_args))
        
    def run_main(self, *args):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        # Required since Jupyter overwrites the signal handler previously set without chaining it.
        _ls_define_signal_handler("SIGINT")
        str_args = _encode_strings_no_null(args)
        _lsp_module_run_main(self._ref, str_args, len(str_args))
        
    def __getitem__(self, var_name):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        result = _lsp_module_get(self._ref, _encode_string_no_null(var_name))
        return _extract_lsp_data(self._modeler, result)

    def __setitem__(self, var_name, value):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        lsp_val = _create_lsp_data(self._modeler, value)
        _lsp_module_set(self._ref, _encode_string_no_null(var_name), lsp_val)

    def __delitem__(self, var_name):
        _check_modeler(self._modeler)
        _check_ref(self._ref)
        _lsp_module_set(self._ref, _encode_string_no_null(var_name), _lspdata())

    def __contains__(self, key):
        return self.__getitem__(key) is not None

########



class LSPModeler(object):
    __slots__ = "_modeler_ptr", "_native_functions", "_attached_solvers", "_reference_scopes", "_stdout", "_stderr", "_stdout_funcs", "_stderr_funcs"

    def __init__(self):
        self._modeler_ptr = _lsp_create_modeler()
        self._native_functions = []
        self._attached_solvers = []
        self._reference_scopes = []
        LSPReferenceScope(self)
        self.set_stdout(sys.stdout)
        self.set_stderr(sys.stderr)

    def delete(self):
        if self._modeler_ptr is None:
            return

        for i in reversed(range(len(self._reference_scopes))):
            self._reference_scopes[i].__del__()

        _lsp_delete_modeler(self._modeler_ptr)
        self._modeler_ptr = None

    def __del__(self):
        self.delete()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.delete()

    def create_solver(self):
        _check_modeler(self)
        self._attached_solvers = [solver for solver in self._attached_solvers if solver._solver_modeler_ptr is not None]
        
        solver_modeler = _lsp_create_solver(self._modeler_ptr).val.ref
        handle_ref = _lsp_solver_handle(solver_modeler).val.ref
        solver_ref = _lsp_handle_solver(handle_ref)

        solver = LocalSolver(solver_ref, handle_ref, solver_modeler)
        self._attached_solvers.append(solver)
        self._reference_scopes[-1]._references.append(solver)
        return solver

    def load_module(self, filepath):
        _check_modeler(self)
        data = _lsp_load_module(self._modeler_ptr, _encode_string_no_null(filepath))
        return LSPModule(self, data)

    def create_module(self, module_name):
        _check_modeler(self)
        data = _lsp_create_module(self._modeler_ptr, _encode_string_no_null(module_name))
        return LSPModule(self, data)

    def create_map(self):
        _check_modeler(self)
        return LSPMap(self, _lsp_create_map(self._modeler_ptr))

    def create_function(self, name, func=None):
        _check_modeler(self)
        if func is None and name is not None:
                func, name = name, func
        arguments = []

        def native_function(_1, args, nb_args, result, _2):
            del arguments[:]
            try:
                for i in range(1, nb_args):
                    arguments.append(_extract_lsp_data(self, args[i]))
                result[0] = _create_lsp_data(self, func(self, *arguments))
            except:
                _ls_set_pending_error(sys.exc_info()[1])
                _ls_interrupt(_encode_string(repr(sys.exc_info()[1])), None)
        
        native_func = _lsp_function_type_2(native_function)
        self._native_functions.append(native_func)
        data = _lsp_create_function_2(self._modeler_ptr, _encode_string_no_null(name), native_func, None)
        return LSPFunction(self, data)

    def get_stdout(self):
        return self._stdout

    def set_stdout(self, stream):
        self._stdout_funcs = self._set_std_stream(1, stream)
        self._stdout = stream

    def get_stderr(self):
        return self._stderr

    def set_stderr(self, stream):
        self._stderr_funcs = self._set_std_stream(2, stream)
        self._stderr = stream

    def _set_std_stream(self, fd, stream):
        def writer_function(_1, content, length, _2):
            try:
                if stream is None:
                    return
                stream.write(_decode_subset_bytes(content, length))
            except:
                _ls_set_pending_error(sys.exc_info()[1])
                _ls_interrupt(_encode_string(repr(sys.exc_info()[1])), None)

        def flusher_function(_1, _2):
            try:
                if stream is None:
                    return
                stream.flush()
            except:
                _ls_set_pending_error(sys.exc_info()[1])
                _ls_interrupt(_encode_string(repr(sys.exc_info()[1])), None)

        writer_func = _lsp_stream_writer_type(writer_function)
        flusher_func = _lsp_stream_flusher_type(flusher_function)
        _lsp_set_std_stream(self._modeler_ptr, fd, writer_func, flusher_func, None)
        return (writer_func, flusher_func)

    def __eq__(self, other):
        return (isinstance(other, LSPModeler)
                and self._modeler_ptr == other._modeler_ptr)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self._modeler_ptr

    stdout = property(get_stdout, set_stdout)
    stderr = property(get_stderr, set_stderr)
########

